import numpy as np
import scipy.stats as stats

class SignalNoiseRatio:
    def __init__(self, debug=False, ):

        self.SNR_COMPATIBLE = False
        self.debug = debug

    def find_index_of_outliers(self, y, z_threshold):

        z_scores = abs(stats.zscore(y))
        filter_arr = z_scores > z_threshold
        indexes_of_outliers = np.where(filter_arr == True)[0]

        return indexes_of_outliers

    def find_index_and_strength_of_outliers(self, signal):
        indexes_of_outliers = self.find_index_of_outliers(signal, 3)

        no_outlier_signal = signal.copy()
        for indx in indexes_of_outliers:
            no_outlier_signal[indx] = 0
        no_outlier_signal_mean = sum(no_outlier_signal) / (len(signal) - len(indexes_of_outliers))
        no_outlier_signal_std = np.std(no_outlier_signal)

        outlier_strengths = []
        for indx in indexes_of_outliers:
            outlier_value = signal[indx]
            outlier_strength = (outlier_value - no_outlier_signal_mean) / no_outlier_signal_std
            outlier_strengths.append(round(outlier_strength, 2))

        recency_of_outliers = [round(i / len(signal), 2) for i in indexes_of_outliers]

        return indexes_of_outliers, recency_of_outliers, outlier_strengths

    def calculate_snr(self, signal):

        indexes_of_outliers, recency_of_outliers, outlier_strengths = self.find_index_and_strength_of_outliers(signal)

        noise_scores = []
        for r, s in zip(recency_of_outliers, outlier_strengths):
            STRENGTH_coeff = 3
            # 3 is used because outlier detection assumes normal distribution and 3 ((signal_point-signal_mean)/ std) is the limit to detect outliers.
            # STRENGTH_coeff=5
            noise = r * ((1 - STRENGTH_coeff / s) ** 0.5 * 5)
            noise_scores.append(round(noise, 2))

        if self.debug == True:
            print("recency_of_outliers:", recency_of_outliers)
            print("strengths:", outlier_strengths)
            print("noise_scores:", noise_scores)

        outliers = signal[indexes_of_outliers].to_list()

        snr = 5 - sum(noise_scores)
        snr = round(snr, 2)
        snr = float(snr)

        snr_details = {
            "indices": indexes_of_outliers,
            "strengths": outlier_strengths,
            "outliers": outliers

        }

        # snr_results=SNRResults(snr,snr_details)

        return snr, snr_details





import pandas as pd
from .signal_to_noise_ratio import SignalNoiseRatio
from time_pattern_cohesion_score import TPCS

# tpcs, scipy, numpy, pandas,

# check_quality requires a signal input pandas pd format. With at least 2 columns where y is where values are stored
# and ds where date information is stores

# END_OF_TIME parameter is useful when we have extra information about time period for the signal. F
# for example signal may have values until May but it should have values for april too.  In this case
# END_OF_TIME param help us determine possible missing values.

# MAX_LEN_MONTHS this parameter works similar way with END_OF_TIME parameter. However it's purpose is to
# gauge the existance of missing values from the beginning of defined data gathering period.

# function returns dict of values.\
#     cleaning_score_data_dict = {
#
#         "TPC_features": TPC_features,
#         "TPC_score": TPC_score,
#         "SNR_features": SNR_features,
#         "SNR_score": SNR_score,
#         "cleaning_score_weights": cleaning_score_weights,
#         "cleaning_score": cleaning_score,
#
#     }
# you can get any value you want, the final output key is "cleaning_score"

def check_quality(signal, MAX_LEN_MONTHS, END_OF_TIME=None, tpcs_limit_for_snr_calculation=3.5):
    assert isinstance(signal, pd.DataFrame)

    assert 'y' in signal.columns
    assert 'ds' in signal.columns

    y = signal["y"]

    x = signal["ds"].to_list()
    x = pd.to_datetime(x)

    # these weights are tested and working well. However you may adjust according to your need
    weights = {"contiguity_score": 0.5,
               "recent_contiguity_score": 3,
               "consistency_score": 0.8,
               "intra_consistency_score": 1

               }


    if END_OF_TIME:
        tpcs = TPCS(x, MAX_LEN_MONTHS, END_OF_TIME=END_OF_TIME, debug=False, return_details=False)
    else:
        last_date=x[-1]
        tpcs = TPCS(x, MAX_LEN_MONTHS, END_OF_TIME=last_date, debug=False, return_details=True)

    TPC_score, tpcs_details = tpcs.calculate_TPCS(weights=weights, printing=False)

    if TPC_score > tpcs_limit_for_snr_calculation:
        snr = SignalNoiseRatio()
        SNR_score, snr_details = snr.calculate_snr(y)
        outlier_count = len(snr_details["outliers"])

    else:

        snr_details = {}
        SNR_score = 0
        outlier_count = None
        snr_details["indices"] = None
        snr_details["strengths"] = None
        snr_details["outliers"] = None

    TPC_features = {'consistency': {'value': tpcs_details["consistency"],
                                    'weight': weights["consistency_score"],
                                    'weighted_value': tpcs_details["weighted_consistency"]
                                    },
                    'intra_consistency': {'value': tpcs_details["intra_consistency"],
                                          'weight': weights["intra_consistency_score"],
                                          'weighted_value': tpcs_details["weighted_intra_consistency"]
                                          },
                    'contiguity': {'value': tpcs_details["contiguity"],
                                   'weight': weights["contiguity_score"],
                                   'weighted_value': tpcs_details["weighted_contiguity"]
                                   },
                    'recent_contiguity': {'value': tpcs_details["recent_contiguity"],
                                          'weight': weights["recent_contiguity_score"],
                                          'weighted_value': tpcs_details["weighted_recent_contiguity"]
                                          }
                    }

    SNR_features = {'outlier_count': outlier_count,
                    'outlier_indices': snr_details["indices"],
                    'outliers': snr_details["outliers"],
                    'outlier_strengths': snr_details["strengths"]}



    TPC_weight = 0.5
    SNR_weight = 0.5

    cleaning_score_weights = {'TPC_weight': TPC_weight,
                              'SNR_weight': SNR_weight
                              }

    cleaning_score = (TPC_score + SNR_score) / 2
    cleaning_score = round(cleaning_score, 2)

    cleaning_score_data_dict = {

        "TPC_features": TPC_features,
        "TPC_score": TPC_score,
        "SNR_features": SNR_features,
        "SNR_score": SNR_score,
        "cleaning_score_weights": cleaning_score_weights,
        "cleaning_score": cleaning_score,

    }

    return cleaning_score_data_dict

